;;; cl-pdf copyright 2002 Marc Battyani see license.txt for the details
;;; You can reach me at marc.battyani@fractalconcept.com or marc@battyani.net

(in-package pdf)

(defconstant +pdf-header+ "%PDF-1.3")

(defvar *document* nil)
(defvar *root-page* nil)
(defvar *page* nil)
(defvar *page-stream* nil)
(defvar *pdf-stream* nil)
(defvar *xrefs* nil)
(defvar *name-counter* 100)
(defconstant +external-format+ #-(or lispworks clisp allegro) :default
   #+(and allegro mswindows) :1252-base
   #+(and allegro unix) :default
   #+lispworks '(:default :eol-style :lf)
   #+clisp :unix)

(defvar *min-size-for-compression* 300)
(defvar *compress-streams* t) ;nil if you don't have or don't want to use zlib

(defun gen-name (prefix)
  (format nil "~a~d" prefix (incf *name-counter*)))

(defclass dictionary ()
  ((dict-values :accessor dict-values :initform nil :initarg :dict-values)))

(defun get-dict-value (dict name)
  (cdr (assoc name (dict-values dict) :test #'string=)))

(defun change-dict-value (dict name value)
  (setf (cdr (assoc name (dict-values dict) :test #'string=)) value))

(defclass pdf-stream (dictionary)
  ((content :accessor content :initform "" :initarg :content)))

(defmethod initialize-instance :after ((obj pdf-stream) &rest init-options &key empty &allow-other-keys)
  (unless empty
    (push `("/Length" . ,#'(lambda () (length (content obj))))
	  (dict-values obj))))

(defclass indirect-object ()
  ((obj-number :accessor obj-number :initform (incf (last-object-number *document*)) :initarg :obj-number)
   (gen-number :accessor gen-number :initform 0 :initarg :gen-number)
   (content :accessor content :initform nil :initarg :content)))

(defmethod initialize-instance :after ((obj indirect-object) &rest init-options &key no-link &allow-other-keys)
  (unless no-link
    (vector-push-extend obj (objects *document*))))

(defclass object-ref ()
  ((obj-number :accessor obj-number :initform 0 :initarg :obj-number)
   (gen-number :accessor gen-number :initform 0 :initarg :gen-number)))

(defclass page-node (indirect-object)
  ((pages :accessor pages :initform (make-array 10 :fill-pointer 0 :adjustable t))))

(defmethod initialize-instance :after ((obj page-node) &rest init-options &key no-link &allow-other-keys)
  (when (and *root-page* (not no-link))
    (vector-push-extend obj (pages *root-page*)))
  (setf (content obj) (make-instance 'dictionary
		       :dict-values `(("/Type" . "/Pages")
				      ("/Count" . ,#'(lambda ()(length (pages obj))))
				      ,@(when *root-page* `(("/Parent" . ,*root-page*)))
				      ("/Kids" . ,(pages obj))))))

(defclass page (indirect-object)
  ((bounds :accessor bounds :initform #(0 0 595 841) :initarg :bounds)
   (resources :accessor resources :initform (make-instance 'dictionary))
   (fonts :accessor fonts :initform (make-instance 'dictionary))
   (xobjects :accessor xobjects :initform (make-instance 'dictionary))
   (content-stream :accessor content-stream)
   ))

(defmethod initialize-instance :after ((page page) &rest init-options &key no-link (rotate 0) &allow-other-keys)
  (when (and *root-page* (not no-link))
    (vector-push-extend page (pages *root-page*)))
  (setf (content-stream page) (make-instance 'pdf-stream))
  (push (cons "/Font" (fonts page)) (dict-values (resources page)))
  (push (cons "/XObject" (xobjects page)) (dict-values (resources page)))
  (let ((content (make-instance 'indirect-object :content (content-stream page))))
    (setf (content page)
	  (make-instance 'dictionary
		:dict-values `(("/Type" . "/Page")
			       ("/Parent" ,*root-page*)
			       ("/MediaBox" . ,#'(lambda ()(bounds page)))
			       ("/Resources" . ,(resources page))
			       ("/Rotate" . ,rotate)
			       ("/Contents" . ,content))))))

(defclass document ()
  ((objects :accessor objects :initform nil)
   (root-page :accessor root-page :initform nil)
   (catalog :accessor catalog :initform nil)
   (last-object-number :accessor last-object-number :initform 0)))

(defmethod initialize-instance :after ((doc document) &rest init-options
				       &key empty &allow-other-keys)
  (unless empty
    (let ((*document* doc))
      (setf (objects doc) (make-array 10 :fill-pointer 0 :adjustable t))
      (setf (catalog doc) (make-instance 'indirect-object))
      (setf (root-page doc) (make-instance 'page-node))
      (setf (content (catalog doc))
	    (make-instance 'dictionary
			   :dict-values `(("/Type" . "/Catalog")
					  ("/Pages" . ,(root-page doc))))))))

(defmacro enforce-/ (&rest names)
  `(progn
    ,@(loop for name in names
	    collect `(unless (char= (schar ,name 0) #\/)
		      (setf ,name (concatenate 'string "/" ,name))))))

(defclass font (indirect-object)
  ((name :accessor name :initform (gen-name "/CLF") :initarg :name)))

(defmethod initialize-instance :after ((font font) &rest init-options &key
				       first-char last-char widths (type "Type1")
				       (base-font "Helvetica")(encoding "WinAnsiEncoding")
				       &allow-other-keys)
  (enforce-/ type base-font encoding)
  (let ((widths-object (when widths (make-instance 'indirect-object :content widths))))
    (setf (content font)
	  (make-instance 'dictionary
             :dict-values `(("/Type" . "/Font")("/Subtype" . ,type)
			    ,@(when first-char `(("/FirstChar" . ,first-char)))
			    ,@(when last-char `(("/LastChar" . ,last-char)))
			    ,@(when widths `(("/Widths" . ,widths-object)))
			    ("/BaseFont" . ,base-font)("/Encoding" . ,encoding))))))

(defun add-fonts-to-page (&rest fonts)
  (dolist (font fonts)
    (push (cons (name font) font) (dict-values (fonts *page*)))))

(defclass image (indirect-object)
  ((name :accessor name :initform (gen-name "/CLI") :initarg :name)))

(defmethod initialize-instance :after ((image image) &rest init-options &key
				       bits width height (filter "ASCIIHexDecode")
				       (color-space "DeviceRGB")(bits-per-color 8)
				       &allow-other-keys)
  (enforce-/ filter color-space)
  (setf (content image)
	(make-instance 'pdf-stream
	       :dict-values `(("/Type" . "/XObject")("/Subtype" . "/Image")
			      ("/Width" . ,width)("/Height" . ,height)
			      ("/Filter" . ,filter)
			      ("/ColorSpace" . ,color-space)
			      ("/BitsPerComponent" . ,bits-per-color))))
  (setf (content (content image)) bits))

(defun add-images-to-page (&rest images)
  (dolist (image images)
    (push (cons (name image) image) (dict-values (xobjects *page*)))))

(defmethod write-object ((obj dictionary) &optional root-level)
  (write-string "<< " *pdf-stream*)
  (loop for (key . val) in (dict-values obj) do
	(write-string key *pdf-stream*)
	(write-char #\Space *pdf-stream*)
	(write-object val)
	(write-char #\Newline *pdf-stream*))
  (write-line " >>" *pdf-stream*))

(defmethod write-object ((obj pdf-stream) &optional root-level)
  #+(or Lispworks allegro CMU)
  (when (and *compress-streams* (> (length (content obj)) *min-size-for-compression*))
    (setf (content obj) (compress-string (content obj)))
    (let ((filter (get-dict-value obj "/Filter")))
      (if filter
	(change-dict-value obj "/Filter" (vector "/FlateDecode" filter))
	(push (cons "/Filter" "/FlateDecode")(dict-values obj)))))
  (call-next-method)
  (write-line "stream" *pdf-stream*)
  (write-sequence (content obj) *pdf-stream*)
  (write-char #\Newline *pdf-stream*)
  (write-line "endstream" *pdf-stream*))

(defmethod write-object ((obj object-ref) &optional root-level)
  (format *pdf-stream* "~d ~d R" (obj-number obj)(gen-number obj)))

(defmethod write-object ((obj indirect-object) &optional root-level)
  (if root-level
    (progn
      (vector-push-extend (format nil "~10,'0d ~5,'0d n "
				  (file-position *pdf-stream*)(gen-number obj))
			  *xrefs*)
      (format *pdf-stream* "~d ~d obj~%" (obj-number obj)(gen-number obj))
      (write-object (content obj))
      (format *pdf-stream* "endobj~%" (obj-number obj)(gen-number obj)))
    (format *pdf-stream* "~d ~d R" (obj-number obj)(gen-number obj))))

(defmethod write-object ((list list) &optional root-level)
  (dolist (obj list)
    (write-object obj)
    (write-char #\Space *pdf-stream*)))

(defmethod write-object ((obj string) &optional root-level)
  (write-string obj *pdf-stream*))

(defmethod write-object ((obj symbol) &optional root-level)
  (write-string (symbol-name obj) *pdf-stream*))

(defmethod write-object ((obj function) &optional root-level)
  (write-object (funcall obj)))

(defmethod write-object ((obj t) &optional root-level)
  (format *pdf-stream* "~a" obj))

(defmethod write-object ((array array) &optional root-level)
  (write-string "[ " *pdf-stream*)
  (loop for obj across array do
	(write-object obj)
	(write-char #\Space *pdf-stream*))
  (write-char #\] *pdf-stream*))

(defmethod write-document ((s stream) &optional (document *document*))
   (let ((*xrefs* (make-array 10 :adjustable t :fill-pointer 0))
	 startxref
	 (*pdf-stream* s))
      (vector-push-extend "0000000000 65535 f " *xrefs*)
      (write-line +pdf-header+ s)
      (loop for obj across (objects document)
	    for first = t then nil
	    if obj do (write-object obj t)
	    else do (unless first (vector-push-extend "0000000000 00001 f " *xrefs*)))
      (setf startxref (file-position s))
      (format *pdf-stream* "xref~%0 ~d~%" (length *xrefs*))
      (loop for xref across *xrefs*
	    do (write-line xref s))
      (format s "trailer~%<< /Size ~d~%/Root " (1- (length (objects document))))
      (write-object (catalog document))
      (format s "~%>>~%startxref~%~d~%%%EOF~%" startxref)))

(defmethod write-document ((filename pathname) &optional (document *document*))
   (with-open-file (s filename :direction :output :if-exists :supersede
		      :external-format +external-format+)
     (write-document s document)))

(defmacro with-document ((&rest args) &body body)
  `(let* ((*document* (make-instance 'document ,@args))
	  (*root-page* (root-page *document*)))
    ,@body))

(defmacro with-page ((&rest args) &body body)
  `(let* ((*page* (make-instance 'page ,@args)))
    (setf (content (content-stream *page*))
     (with-output-to-string (*page-stream*)
			    ,@body))))
