;;;
;;; Copyright (c) 2002 Edi Weitz <edi@agharta.de>
;;;
;;; zlib-cmucl.lisp - Code to compress a Lisp string to an array of
;;; bytes, using the zlib compression library (http://www.zlib.org). 
;;;
;;; Usage:
;;;
;;; 1. Set *libgz-path* to the location of the zlib library on your
;;; machine. Compile and load this file
;;;
;;; 2. Call (COMPRESS-STRING "your string"). Two values are returned:
;;; the array of bytes containing the compressed data, and the
;;; actual number of bytes in the compressed output. The length of the 
;;; resulting array is a fixed function of the length of the input
;;; string, while the length of the compressed data is variable and
;;; depends on the randomness of the input (strings with a lot of
;;; repetitions will compress better).
;;;

(in-package :pdf)

#-CMU (error "This code should be loaded in CMUCL.")

(use-package "ALIEN")
(use-package "C-CALL")

;; This function stolen from the CMUCL documentation
(defun array-data-address (array)
  "Return the physical address of where the actual data of an array is
stored.

ARRAY must be a specialized array type in CMUCL.  This means ARRAY
must be an array of one of the following types:

                  double-float
                  single-float
                  (unsigned-byte 32)
                  (unsigned-byte 16)
                  (unsigned-byte  8)
                  (signed-byte 32)
                  (signed-byte 16)
                  (signed-byte  8)
"
  (declare (type (or #+signed-array (array (signed-byte 8))
                     #+signed-array (array (signed-byte 16))
                     #+signed-array (array (signed-byte 32))
                     (array (unsigned-byte 8))
                     (array (unsigned-byte 16))
                     (array (unsigned-byte 32))
                     (array single-float)
                     (array double-float))
                 array)
           (optimize (speed 3) (safety 0))
           (ext:optimize-interface (safety 3)))
  ;; with-array-data will get us to the actual data.  However, because
  ;; the array could have been displaced, we need to know where the
  ;; data starts.
  (lisp::with-array-data ((data array)
                          (start)
                          (end))
    (declare (ignore end))
    ;; DATA is a specialized simple-array.  Memory is laid out like this:
    ;;
    ;;   byte offset    Value
    ;;        0         type code (should be 70 for double-float vector)
    ;;        4         4 * number of elements in vector
    ;;        8         1st element of vector
    ;;      ...         ...
    ;;
    (let ((addr (+ 8 (logandc1 7 (kernel:get-lisp-obj-address data))))
          (type-size (let ((type (array-element-type data)))
                       (cond ((or (equal type '(signed-byte 8))
                                  (equal type '(unsigned-byte 8)))
                              1)
                             ((or (equal type '(signed-byte 16))
                                  (equal type '(unsigned-byte 16)))
                              2)
                             ((or (equal type '(signed-byte 32))
                                  (equal type '(unsigned-byte 32)))
                              4)
                             ((equal type 'single-float)
                              4)
                             ((equal type 'double-float)
                              8)
                             (t
                              (error "Unknown specialized array element type"))))))
      (declare (type (unsigned-byte 32) addr)
               (optimize (speed 3) (safety 0) (ext:inhibit-warnings 3)))
      (system:int-sap (the (unsigned-byte 32)
                        (+ addr (* type-size start)))))))

(defvar *libgz-path* "/lib/libz.so.1"
  "Set this variable to point to the location of the zlib library on
your system.")

(defvar *libgz-loaded* nil)

(unless *libgz-loaded*
  (load-foreign *libgz-path*):q
  (setq *libgz-loaded* t))

(declaim (inline gz-string))
(def-alien-routine ("compress" gz-string)
    integer
  (dest (* (unsigned 8))) 
  (destlen long :in-out)
  (source c-string)
  (sourcelen long))

(defun compress-string (source)
  "Compress the string SOURCE. Returns two values: the array of bytes
representing the compressed data and the number of compressed bytes."
  (let* ((sourcelen (length source))
	 (destlen (+ 12 (ceiling (* sourcelen 1.05))))
	 (dest (make-array destlen
                           :element-type '(unsigned-byte 8)
			   :initial-element 0)))
    (multiple-value-bind (res new-destlen)
        (gz-string (array-data-address dest)
                   destlen source sourcelen)
      (if (zerop res)
          (values (map 'string #'code-char (subseq dest 0 new-destlen))
                  new-destlen)
          (error "zlib error, code ~d" res)))))
